;----- A Forth interpreter/compiler in m4 that cross-compiles to LLL -----


(provide 'forth)

;------------------------------ Module parameters ----------------------------

This module defines the meta- part of a FORTH meta-compiler,
with hooks for actual compilation words,
and m4 macros to define basic FORTH words in other modules.

One uses the FORTH meta-compiler by calling the __FORTH__ macro
with the text to interpret as parameter.

All parameter and provided macros are of the form __FORTH__$1__.
Here are their meaning according to $1:


------ Parameters ------

--- Compiling words:
COMPILE_DATA_PUSH	compiles the pushing of a literal in $1 on data stack.





------ Provided words ------

--- variables
RUNTIME_SEMANTICS	run-time semantics being defined

--- commands



------ Provided FORTH words ------
I tried to be ANSI compliant as to what these words do



------ Internals ------

--- variables
INPUT			input buffer
OUTPUT			output command
WORD			last parsed word
SEARCH_ORDER		comma-separated search order of meta-vocabularies
WORD_SEMANTICS		found semantics for the word
INTERP_SEMANTICS	interpretation semantics being defined
COMPILE_SEMANTICS	compile semantics being defined
VOCABULARY_CURRENT	current vocabulary

--- flags
IS_NUM			is last parsed word a literal ?
EOF			was end of buffer reached during parsing ?
WORD_FOUND		was the word found in the search order

--- commands
PLEASE			feeds the Forth input with arguments.
ERROR			report an error (default output command)
GET_TOKEN		reads token and analyzes it
GET_WORD		(destructively) reads a token from INPUT to WORD
ANALYZE_TOKEN		analyzes a word, looking for a literal
DATA_PUSH		pushes data on the (meta) data stack
VOCABULARY_MAKE		makes a new meta-vocabulary
VOCABULARY_ALSO		adds to the current vocabulary
VOCABULARY_NOMORE	pops from the vocabulary stack


--- continuations
INTERP			the interpreter
COMPILE			the compiler
DO_INTERP		interprets last parsed word
DO_COMPILE		compiles last parsed word


--- structures
INTERP_SEMANTICS_$$$	indicates interpretation semantics for $$$
COMPILE_SEMANTICS_$$$	indicates compile semantics for $$$

}
-------------------------------- Main loops ---------------------------------

--- The main FORTH word

define({__FORTH__},{
  __FORTH__PLEASE__({$*})
  __FORTH__INTERP__
})


--- Feeding the FORTH input

define({__FORTH__PLEASE__},{PrependVar({__FORTH__INPUT__},{$*})})



--- Forth Interpreter continuation 

define({__FORTH__INTERP__},{
   __FORTH__GET_TOKEN__
   __FORTH__DO_INTERP__
})


--- Forth compiler continuation

define({__FORTH__COMPILE__},{
   __FORTH__GET_TOKEN__
   __FORTH__DO_COMPILE__
})


--- Grabbing one token from the Forth input

define({__FORTH__GET_TOKEN__},{
   __FORTH__GET_WORD__
   __FORTH__ANALYZE_TOKEN__
})


--- report an error
define({__FORTH__ERROR__},{_OUTln_({$1})})


--- Forth output
define({__FORTH__OUTPUT__},{__FORTH__ERROR__({$1})})



--- Grabbing one word from the Forth input

define({__FORTH__GET_WORD__},{
   regexp(defn({__FORTH__INPUT__},
          \`__REGEXP_BLANK__*\(__REGEXP_NOBLANK__*\)\(__REGEXP_ANY__*\)\',
	  {define({__FORTH__WORD__},{\1})
	   define({__FORTH__INPUT__},{\2})
	  })
})


--- Analyzing an input word into a token
can only be an integer literal or a word

define({__FORTH__ANALYZE_TOKEN__},{
   define({__FORTH__IS_NUM__},
   	ifelse(regexp(defn({__FORTH__WORD__}),
			\`\([0-9]+\|0x[0-9a-bA-B]+\)\'),
	       -1,0,1))
   define({__FORTH__EOF__},ifelse(defn({__FORTH__WORD__}),{},1,0))
})


--- Interpret a token

define({__FORTH__DO_INTERP__},{
   __FORTH__SEARCH__({__FORTH__INTERP_SEM__})
   ifelse(__FORTH__WORD_FOUND__,1,{__FORTH__WORD_SEM__},
	  {__FORTH__DATA_PUSH__(__FORTH_WORD__)})
})


--- Compile a token

define({__FORTH__DO_COMPILE__},{
   __FORTH__SEARCH__({COMPILE_SEMANTICS})
   ifelse(__FORTH__WORD_FOUND__,1,{__FORTH__WORD_SEMANTICS__},
	  {__FORTH__COMPILE_DATA_PUSH__(__FORTH_WORD__)})
})



------------------------------ meta-vocabularies ----------------------------
Create the vocabulary:
Make it first in the search order:
define({__FORTH__VOCABULARY_MAKE__},{
   __FORTH__VOCABULARY_ALSO__({$1})
})


Put vocabulary first in search order:
Put it as current vocabulary target:
define({__FORTH__VOCABULARY_ALSO__},{
   pushdef({__FORTH__SEARCH_ORDER__})
   PrependList({__FORTH__SEARCH_ORDER__},{$1})
   pushdef({VOCABULARY_CURRENT},{$1})
})

Pop a vocabulary from the list
define({__FORTH__VOCABULARY_NOMORE__},{
   popdef({__FORTH__SEARCH_ORDER__})
   popdef({VOCABULARY_CURRENT})
})


__FORTH__VOCABULARY_MAKE__({M4FORTH})


define({__FORTH__SEARCH__},....)

define({__FORTH__DEFINE__},.....)



--- The Forth meta data stack
__newstk__(FORTH_DATA)
define({__FORTH__DATA_PUSH__},{__push__({FORTH_DATA},{$1})})
define({__FORTH__DATA_POP__},{__pop__({FORTH_DATA})})
define({__FORTH__DATA_PICK__},{__pick__({FORTH_DATA},{$1})})


--- The Forth meta return stack
__newstk__(FORTH_RETURN)
define({__FORTH__RETURN_PUSH__},{__push__({FORTH_RETURN},{$1})})
define({__FORTH__RETURN_POP__},{__pop__({FORTH_RETURN})})
define({__FORTH__RETURN_PICK__},{__pick__({FORTH_RETURN},{$1})})



--- The Forth object RETURN stack

define({__FORTH__COMPILE_DATA_PUSH__},.....)



